;; Generated package description from yaml.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "yaml" "1.2.4.0.20260605.0" "YAML parser for Elisp" '((emacs "25.1")) :commit "5546f36bde24a9a8c1934e0f6ce205cd41d72537" :authors '(("Zachary Romero" . "zkry@posteo.org")) :maintainer '("Zachary Romero" . "zkry@posteo.org") :keywords '("tools") :url "https://github.com/zkry/yaml.el")
